/******************************************************************************
 * ProblemSolver.h
 *
 * GDSAT
 *
 ******************************************************************************
 * Copyright (C) 2013 Benjamin Niedermann <niedermann@kit.edu>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program.  If not, see <http://www.gnu.org/licenses/>.
 *****************************************************************************/

#ifndef PROBLEMSOLVER_H_
#define PROBLEMSOLVER_H_

#include "Model.h"
#include "SATSolver.h"
#include "Graph.h"
#include "Clock.h"

typedef pair<int,int> ClosedInterval;
struct Config;
class  ProblemModel;
struct Result;


/**
 * Wraps the process of searching a solution for a given problem parameter, e.g., pathwidth.
 * APPROACH:
 * do{
 *   1.) based on a given graph create a model for the i-th iteration, e.g., test whether graph has pathwidth i.
 *   2.) apply the sat solver on model.
 *   3.) if model is satisfied return solution
 * }while(not last iteration) (nextTurn(...))
 * return that no solution could be found.
 */
class ProblemSolver {
public:
	ProblemSolver();
	virtual ~ProblemSolver();


	/**
	 * For a given graph and problem parameter (e.g. pathwidth) this method searches for a solution by iteratively
	 * testing values in increasing order for the problem parameter.
	 * The problem parameter is specified by the subclasses implementing the virtual methods.
	 * - graph: the problem solver is applied on this graph.
	 * - solution: the obtained solution: return by parameter
	 * - config: Configuration of the problem solver. This parameter can be used to configure the specific problem solver.
	 * (see subclasses)
	 *
	 *STRUCTURE:
	 * do{
	 *   1.) create model for i-th iteration (createModel(...))
	 *   2.) apply sat solver on model.
	 *   3.) if model is satisfied return solution (analyze(...))
	 * }while(not last iteration) (nextTurn(...))
	 * return that no solution could be found.
	 */
	void execute(Graph & graph, Result & solution, Config * config);

	/**
	 * creates the model for the i-th iteration.
	 * - turn: current iteration: can be used to determine the problem parameter, e.g., pathwidth.
	 */
	virtual ProblemModel * createModel(Graph & graph, Clock & clock, int turn, Config * config) =0;
	/**
	 * analyzes the solution obtained in step 2.) of method execute(). If it returns true the procedure
	 * execute is aborted and it is assumed that the problem could be solved.
	 */
	virtual bool           analyze(SolverResult & result, ProblemModel * model, Result & solution)=0;

	/**
	 * Returns true if a further iteration in method execute should be applied on the graph. This method
	 * should return false if the upper bound is exceeded.
	 */
	virtual bool           nextTurn(Graph & graph,int currentTurn, SolverResult & result)=0;
};


/**
 * Stores the CNF model created by the ProblemSolver. Subclasses of this class also store the
 * interpretation of the CNF Model, i.e., the relation between nodes, edges and variables.
 */
class ProblemModel{
    private:
	CNFModel * cnfModel;

    public:
	CNFModel & getCNFModel(){return *cnfModel;}

	ProblemModel(int varCount, int cnfCount){
		cnfModel = new CNFModel(varCount,cnfCount);
	}
	virtual ~ProblemModel(){
		delete cnfModel;
	}

	/**
	 * Assuming that the underlying cnf model is satisfied this methods
	 * applies some sanity checks. In case that such a check fails the
	 * method throws an exception.
	 */
	virtual void checkForFeasibility(Graph & graph) =0;
	virtual int  getValue()=0;

	/**
	 * For a range of variables this method returns all positions of variables
	 * that are set true.
	 * The call of the method requires:
	 * 1.) the variables that are set true form a interval.
	 * 2.) the CNF model is already solved.
	 */
	ClosedInterval getInterval(Range & variables);
};


/**
 * Stores the result of the execution of a ProblemSolver instance.
 */
struct Result{
	SolverResult status;
	ProblemModel * model;
	/**
	 * time for creating the model.
	 */
    Clock preClock;
    /**
     * time for executing the SAT solver.
     */
    Clock solverClock;
    int getValue(){
    	return model->getValue();
    }
};

/**
 * Subclasses of this class describe the configuration
 */
struct Config{};


#endif /* PROBLEMSOLVER_H_ */
