/******************************************************************************
 * VisibilityRepresentation.cpp
 *
 * GDSAT
 *
 ******************************************************************************
 * Copyright (C) 2013 Benjamin Niedermann <niedermann@kit.edu>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program.  If not, see <http://www.gnu.org/licenses/>.
 *****************************************************************************/
#include <sstream>
#include <iostream>
#include <fstream>
#include <stdlib.h>
#include <sys/time.h>
#include <set>
#include "VisibilityRepresentation.h"
#include "Grid.h"
#include "./../common/Model.h"
#include "./../common/Clause.h"
#include "./../common/SATSolver.h"
#include "./../common/Box.h"


VRModel * createVRModel(int width, int height,bool oneBar, Graph & graph, Clock& clock){


	clock.start();
	Box * nodes = new Box[graph.n()];
	Box * edges = new Box[graph.m()];
	Grid *  neighbors = new Grid[graph.m()*2];
	height +=2;
	width +=2;
	int length = width*height;
	int * size = new int[2];
	size[0] = width;
	size[1] = height;

	Grid * oneBarVars;

	if(oneBar){
		oneBarVars = new Grid[graph.m()];
	}


	int sizeNodesAreDisjoint = (graph.n()*(graph.n()-1)/2)*length;
	int sizeEdgesAreDisjoint = length*graph.m()*graph.m();

    int sizeEdgesAreSegments = width*graph.m()+height*graph.n();
    int sizeIncidentEdgesAndNodes = graph.m()*2*length*2+graph.m()*2;

    int sizeOneBar = oneBar ? graph.m()*length*(length-1)/2 : 0;

	int sizeEdgesAndNodesAreDisjoint =0;


	for(int i=0; i < graph.m(); ++i){
		for( int j=0; j < graph.n(); ++j){
			pair<int, int> edge = graph.edges()[i];
			if(j != (unsigned int)edge.first && j != (unsigned int)edge.second){
				sizeEdgesAndNodesAreDisjoint++;
			}
		}
	}

	sizeEdgesAndNodesAreDisjoint =sizeEdgesAndNodesAreDisjoint*length;

	int cnfCount = sizeEdgesAndNodesAreDisjoint
			      + sizeEdgesAreDisjoint+sizeEdgesAreSegments
			      + sizeIncidentEdgesAndNodes
			      + sizeNodesAreDisjoint
			      + Box::cnfCount(2,size)*(graph.n()+graph.m())+sizeOneBar;

	int oneBarVarCount = oneBar ? graph.m()*width*height : 0;

	int varCount = Box::varCount(2,size)*(graph.n()+graph.m())+oneBarVarCount+graph.m()*2*length;

	VRModel * vrModel = new VRModel(varCount,cnfCount);

	CNFModel & model = vrModel->getCNFModel();


	vrModel->edges =edges;
	vrModel->nodes = nodes;
	vrModel->height = height;
	vrModel->width = width;
	vrModel->oneBar = oneBar;
	vrModel->neighbors = neighbors;




	for(int i=0; i < graph.n(); ++i){
		nodes[i].init(model,size,2);
	}

	for(int i=0; i < graph.m(); ++i){
		edges[i].init(model,size,2);
	}






	for(int i=0; i < graph.m()*2; ++i){
		neighbors[i].init(width,height,model);

	}

	for(int i=0; i < graph.m() && oneBar; ++i){
		oneBarVars[i].init(width,height,model);
	}






	/** Nodes are disjoint **/


	for( int i=0; i < graph.n(); ++i){
		for(int j=i+1; j < graph.n(); ++j){
				int  x = nodes[i].get();
				int  y = nodes[j].get();

				for(int l=0; l < length; ++l,++x,++y){
					Clause * cnf = model.next(2);
					cnf->add(x,false);
					cnf->add(y,false);
				}
		}
	}





	for( int i=0; i < graph.m(); ++i){
		/** Edges are disjoint **/

		for(int j=0; j < graph.m(); j++){
			if(i != j){
				int x1 = edges[i].get();
				int x2 = edges[j].get();
				for(int pos=0; pos < length; ++pos,++x1,++x2){
					Clause * cnf = model.next(4);
					int e = edges[i].getE(pos,1);
					int b = edges[i].getB(pos,1);
					cnf->add(x1,false);
					cnf->add(x2,false);
					cnf->add(e,true);
					cnf->add(b,true);
				}

			}
		}
	}



	/**  Edges and non-incident nodes are disjoint **/
	 for( int i=0; i < graph.m(); ++i){
			pair<int, int> edge = graph.edges()[i];


			for( int j=0; j < graph.n(); ++j){
				if(j != (unsigned int)edge.first && j != (unsigned int)edge.second){
					int  x1 = edges[i].get();
					int  x2 = nodes[j].get();
					int  y  = oneBar ? oneBarVars[i].get() : 0;
					for(int l=0; l < length; ++l,++x1,++x2,++y){
						Clause * cnf =model.next(oneBar ? 3 :2);
						cnf->add(x1,false);
						cnf->add(x2,false);
						if(oneBar){
							cnf->add(y,true);
						}


					}

				}
			}

			int  y  = oneBar ? oneBarVars[i].get() : 0;
			for(int l=0; l < width*height && oneBar; ++l,++y){
				int  y2  =oneBarVars[i].get()+l+1;
				for(int k=l+1; k < width*height; ++k,++y2){
					Clause * cnf = model.next(2);
					cnf->add(y,false);
					cnf->add(y2,false);

				}
			}

	}





	/** Edges are segments **/

	// edges are vertical segments

	for(int i=0; i < graph.m(); ++i){
		int e = edges[i].beginE(0);
		for(int b = edges[i].beginB(0); b < edges[i].endB(0); ++b,++e ){
			Clause * cnf = model.next(2);
			cnf->add(b,false);
			cnf->add(e,true);
		}
	}

	// nodes are horizontal segments

	for(int i=0; i < graph.n(); ++i){
		int e = nodes[i].beginE(1);
		for(int b = nodes[i].beginB(1); b < nodes[i].endB(1); ++b,++e ){
			Clause * cnf = model.next(2);
			cnf->add(b,false);
			cnf->add(e,true);
		}
	}





	int index=0;
	for(int i=0; i < graph.m()*2; ++i,++index){

			int edge = index/2;
			int node = index%2 ==0 ? graph.edges()[edge].first : graph.edges()[edge].second;

			Grid grid = neighbors[i];


			Box gridE = edges[edge];
			Box gridN = nodes[node];


			Clause * cnf = model.next(length);

			int counter =0;

			for(int row =0; row < height; ++row){
				int x = grid.get(row);
				int xe = gridE.get(row);
				int xn = gridN.get(row);

				for(int col=0; col < width; ++col,++x,++xe,++xn){
					counter++;

					cnf->add(x,true);


					Clause * cnf2 = model.next(2);
					cnf2->add(x,false);
					cnf2->add(xe,true);

					cnf2 = model.next(2);
					cnf2->add(x,false);
					cnf2->add(xn,true);
				}
			}
		}





	clock.stop();

	return vrModel;


}

void VRModel::print(Graph & graph){
	cout << "NODES: " << endl;
	for(int i=0; i< graph.n(); ++i){
		cout <<"Node "<<i<< endl<< nodes[i].toString(getCNFModel()) << endl;
	}
	cout << "EDGES: " << endl;
	for(int i=0; i< graph.m(); ++i){
		cout <<"Edge "<<i<< endl << edges[i].toString(getCNFModel()) << endl;
	}
}


void VRModel::checkForFeasibility(Graph & graph){

	print(graph);

	int noCoveredNodes = oneBar ? 3 :2;

	int  * nodeField = new int[width*height];

	for(int i=0; i < width*height; ++i){
		nodeField[i]=-1;
	}

	for(int i=0; i < graph.n(); ++i){
		int x = nodes[i].get();
		int row=-1;
		for(int pos =0; pos < width*height; ++pos,++x){
			if(getCNFModel().getValue(x)){
				int r = pos/width;
				if(row == -1 ){
					row = r;
				}else if(row != r){
					cout << "node is not a horizontal segment" << endl;
					throw "node is not a horizontal segment";
				}
				if(nodeField[pos] == -1){
					nodeField[pos] = i;
				}else{
					throw "nodes are not disjoint.";
				}
			}
		}
	}

	for(int i=0; i < graph.m(); ++i){
		set<int> coveredNodes;
		int x = edges[i].get();
		pair<int,int> edge = graph.edges()[i];
		int col = -1;
		for(int pos =0; pos < width*height; ++pos,++x){

			if(getCNFModel().getValue(x)){
				int c = pos%width;
				if(col == -1 ){
					col = c;
				}else if(col != c){
					cout << "edge is not a vertical segment" << endl;
					throw "node is not a vertical segment";
				}
				if(nodeField[pos]!=-1){
					coveredNodes.insert(nodeField[pos]);
				}
			}
		}
		if((int)coveredNodes.size() > noCoveredNodes){
				throw "Too many nodes are covered.";
			}
		bool first = false;
		bool second = false;

		for(set<int>::iterator it = coveredNodes.begin(); it != coveredNodes.end(); ++it){
				if((*it) == edge.first) first =true;
				if((*it) == edge.second) second =true;
	    }
		if(!first || !second){
			throw "incident node is not covered.";
		}



	}

}






