/******************************************************************************
 * main.cpp
 *
 * GDSAT
 *
 ******************************************************************************
 * Copyright (C) 2013 Benjamin Niedermann <niedermann@kit.edu>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program.  If not, see <http://www.gnu.org/licenses/>.
 *****************************************************************************/

#include <sys/time.h>
#include <sstream>
#include <iostream>
#include <algorithm>
#include <list>
#include <fstream>
#include <iomanip>
#include "../common/Graph.h"
#include "../common/Clock.h"
#include "../common/SATSolver.h"
#include "../sat1d/Bandwidth.h"
#include "../sat1d/STOrientation.h"
#include "../sat1d/PathWidth.h"
#include "../sat2d/VisibilityRepresentation.h"

#include "../sat2d/Boxicity.h"


#include "../common/ProblemSolver.h"
#include "utils.h"




struct Parameters{
	string command;
	string solverCommand;
	string tempFile;
	string inputFile;
	string outputFile;
	string additionalArgs;
	string problemSolver;
	bool  checkResult;


	Parameters(const char * args[], int count){
		if(count < 2 || count % 2!= 0){
			throw "wrong parameters";
		}
		checkResult = false;
		command = args[1];
		for(int i=2; i < count; i+=2){
			string type = args[i];
			string value = args[i+1];
			if(type == "-t"){tempFile = value;}
			if(type == "-sc"){solverCommand = value;}
			if(type == "-i"){inputFile = value;}
			if(type == "-o"){outputFile = value;}
			if(type == "-a"){additionalArgs = value;}
			if(type == "-ps"){problemSolver = value;}
			if(type == "-check"){checkResult = value.compare("true") == 0;}
		}

	}

	string toString(){
		string s;
		s += "command " + command +"\n";
		s += "-t " + tempFile +"\n";
		s += "-sc " + solverCommand +"\n";
		s += "-i " + inputFile +"\n";
		s += "-o " + outputFile +"\n";
		s += "-a " + additionalArgs +"\n";
		s += "-ps " + problemSolver +"\n";
		s += "-check ";
		s += (checkResult ? "true" : "false");
		s+=  +"\n";
		return s;
	}
};







void runTests();

void extract(string & input, string & output);
void execute(Parameters & parameters);
void chooseRandomNodes(string & input, string & output);


inline bool equal(const char * str1, const char* str2){
	return string(str1).compare(str2) ==0;
}

inline bool equal(string &str1, const char* str2){
	return str1.compare(str2) ==0;
}

void printHelp(){

	cout << "USAGE: sat COMMAND [OPTION]..."<<endl;
	cout << "Executes COMMAND using the parameters given by OPTION."<<endl<<endl;
	cout << "COMMAND can be one of the following two commands:"<<endl;
	cout << "1.) execute"<<endl<<"Given a graph and a graph parameter the strategy searches for the smallest value\nthat satisfies"
			" the graph parameter for the graph. The following parameters must be specified:"<<endl;
			cout << "-i inputFile"<<endl;
			cout << "\t File containing the graph."<<endl;
		    cout << "-ps problemSolver"<<endl;
		    cout << "\t Graph parameter to be considered:"<<endl;
		    cout << "\t\t pws \t pathwidth"<<endl;
		    cout << "\t\t bws \t bandwidth"<<endl;
		    cout << "\t\t os \t st-orientation"<<endl;
		    cout << "\t\t vrs \t bar visibility representation"<<endl;
		    cout << "\t\t bovs \t bar one-visibility representation"<<endl;
		    cout << "-o outputFile"<<endl;
		    cout << "\t	The result of the execution is added to this file."<<endl;
		    cout << "-sc solverCommand"<< endl;
		    cout << "\t	Path of the SAT solver to be executed. So far only minisat is supported."<<endl;
		    cout << "-a additionalParameters"<<endl;
		    cout << "\t Additional parameters that are passed over to the SAT solver."<<endl;
		    cout << "-t tempFile"<<endl;
		    cout << "\t Temporary file that is used for communication with the SAT solver."<<endl;
		    cout << "-check [true|false]"<<endl;
		    cout << "If true then some sanity checks are applied in the end." <<endl<<endl;

	cout << "2.) test"<<endl<<"Executes some tests for checking the implementation."<<endl<< "The parameters -sc and -t must be "
			"specified in the same way as for the previous command."<<endl<<endl;

	cout << "Example:"<<endl<<endl;
	cout << "\t sat execute -ps pws -i /home/graph/rome1.graph -o /home/graph/result.csv -sc /home/programs/minisat/core/minisat "
			"-a \"-cpu-lim=300 -mem-lim=100000000000\" -t /home/temp/tempFile -check true"<<endl<<endl;




}


int main(int argc, const char* argv[]){



	if(argc < 2 || equal(argv[1],"--help")){
		printHelp();
		return 1;
	}
	Parameters pars = Parameters(argv,argc);
	cout << "started with following parameters:"<<endl;
	cout << pars.toString() << endl;
	if(equal(pars.command,"st")){
		cout << "CHOOSE st-Nodes" << endl;
		chooseRandomNodes(pars.inputFile,pars.outputFile);
		return 1;
	}

	if(equal(pars.command,"test")){
		cout << "RUN TESTS" << endl;
		SATSolver::command = pars.solverCommand;
		SATSolver::tempFile = pars.tempFile;
		if(SATSolver::command == "" || SATSolver::command == ""){
			cout << "Parameters for solver have not been set.";
			printHelp();
			return 0;
		}
		cout << SATSolver::command << endl;
		runTests();
		return 1;
	}

	if(equal(pars.command,"statistics") && argc == 4){
		cout << "EXTRACT STATISTICS" << endl;
		extract(pars.inputFile,pars.outputFile);
		return 1;
	}

	if(equal(pars.command,"execute")){
		try{
			SATSolver::command = pars.solverCommand;
			SATSolver::tempFile = pars.tempFile;
			if(SATSolver::command == "" || SATSolver::command == ""){
				cout << "Parameters for solver have not been set.";
				printHelp();
				return 0;
			}
			execute(pars);
		}catch(const char * &error){
			cout << "ERROR: " << error;
			return 0;
		}
		return 1;
	}


	printHelp();
}


void print(fstream & output, Clock& preClock, Clock& solverClock, int itCounter){
	output<< " " << itCounter;
	output<< " " <<(preClock.getOverallTime()+solverClock.getOverallTime()) << " " << preClock.getOverallTime() << " " << solverClock.getOverallTime();
	preClock.write(output);
	solverClock.write(output);
	output<< endl;
}

string print(SolverResult result){
	if(result == SAT) return "SAT";
	if(result == UNSAT) return "UNKNOWN";
	if(result == TIME_OUT) return "TIME_OUT";
	if(result == NOT_SOLVED) return "NOT_SOLVED";
	throw "unspecified result";
}

void execute(ProblemSolver & solver, Graph & graph, Config * config,string & inputFile, fstream & output, bool check ){
	Result result;
	solver.execute(graph,result,config);
	if(result.status == SAT && check){
		result.model->checkForFeasibility(graph);
	}
	double pc = result.preClock.getOverallTime();
	double sc = result.solverClock.getOverallTime();
	output << inputFile <<" "<<graph.n() << " " << graph.m() << " "<<graph.n()+graph.m()
		   << " " << print(result.status) << " " << result.getValue() << " " << (pc+sc) << " "<<pc <<" "<< sc << endl;
}

void execute(Parameters & pars){

	list<pair<int, int> > stPairs;
	Graph graph = Graph();



	fstream file;
	file.open(pars.inputFile.c_str(),ios::in);
	if(file.is_open()){
		readFile(graph,stPairs,file);
	}else{
		cout << "could not open file" << endl;
		throw "could not open file";
	}



	fstream outputFile;
	outputFile.open(pars.outputFile.c_str(),ios::app|ios::out);
	if(!outputFile.is_open()){
		throw "could not open output file";
	}



	if(pars.problemSolver.compare("os")==0){
		for(list<pair<int,int> >::iterator it = stPairs.begin(); it != stPairs.end(); ++it){
			STOConfig config = STOConfig((*it).first,(*it).second);
			STOrientationSolver stSolver;
			execute(stSolver,graph,&config,pars.inputFile,outputFile,pars.checkResult);
			outputFile.flush();
		}

	}else {
		Config * empty =0;


		if(pars.problemSolver.compare("bws")==0){
			cout << "EXECUTE BWS. " << graph.n() << " " << graph.m() <<  endl;
			for(int i=0; i < graph.m(); ++i){
				cout << i <<"=("<<graph.edges()[i].first << ","<<graph.edges()[i].second<<")"<<endl;
			}
			BandwidthSolver bws;
			execute(bws,graph,empty,pars.inputFile,outputFile,pars.checkResult);
		}else if(pars.problemSolver.compare("pws")==0){
			PWSolver pws;
            execute(pws,graph,empty,pars.inputFile,outputFile,pars.checkResult);
		}else if(pars.problemSolver.compare("vrs")==0){
			VRSolver vrs;
			VRConfig config = VRConfig(false);
			execute(vrs,graph,&config,pars.inputFile,outputFile,pars.checkResult);
		}else if(pars.problemSolver.compare("bovs")==0){
			VRSolver obvs;
			VRConfig config = VRConfig(true);
			execute(obvs,graph,&config,pars.inputFile,outputFile,pars.checkResult);
		}else if(pars.problemSolver.compare("bcs")==0){
			BCSolver bcs;
            execute(bcs,graph,empty,pars.inputFile,outputFile,pars.checkResult);
		}
	}
	outputFile.close();
	cout << "##### FINISH #####"<< endl;

}


void chooseRandomNodes(string & inputFile, string & outputFile){

	list<pair<int, int> > stPairs;
	Graph graph = Graph();
	fstream file;
	file.open(inputFile.c_str(),ios::in);
	if(file.is_open()){
		readFile(graph,stPairs,file);
		file.close();
	}else{
		throw "could not open file";
	}


	srand (time(NULL));
    int s = rand() % graph.n();
    int t;
    do{
    	t = rand() % graph.n();

    }while(s==t);

	file.open(outputFile.c_str(),ios::out);
	if(file.is_open()){
		file << "nodeCount " << graph.n()<<endl;
		file << "edges ";
		for(int i=0; i < graph.m(); ++i){
			pair<int,int> edge = graph.edges()[i];
			file << edge.first << " " << edge.second;
			if(i < graph.m()-1){
				file << " ";
			}
		}
		file << endl;
		for(list<pair<int,int> >::iterator it = stPairs.begin(); it != stPairs.end(); ++it){
			file << "st " << (*it).first << " " << (*it).second << endl;
		}
		file << "st " << s<< " " << t << endl;

		file.close();
	}else{
		throw "could not open file";
	}
}

void extract(string & input, string& output){
	list<pair<int, int> > stPairs;
	Graph graph = Graph();
	fstream file;
	file.open(input.c_str(),ios::in);
	if(file.is_open()){
		readFile(graph,stPairs,file);
		file.close();
	}else{
		cout << "could not open file" << endl;
		throw "could not open file";
	}
	cout << "WRITE" << endl;
	fstream out;
	out.open(output.c_str(), ios::out|ios::app);
	out << graph.n() <<" "<< graph.m() <<" "<<graph.m()+graph.n()<<endl;
	out.close();
}



